from datetime import datetime
from cy_components.helpers.formatter import DateFormatter
from cy_components.utils.notifier import MessageHandler, MessageType
from cy_widgets.logger.base import RecorderBase


class ProcedureRecorder(RecorderBase):

    def __init__(self, m_token: str, m_type: MessageType):
        self.__message_token = m_token
        self.__message_type = m_type
        self.__summary_log = ""

    def record_simple_info(self, content):
        print(content, end='\n\n')

    def record_procedure(self, content):
        content = DateFormatter.now_date_string('[%H:%M:%S] ') + content
        print(content, end='\n\n')

    def record_exception(self, content):
        print(content, end='\n\n')
        MessageHandler.send_message(content, 'Precedure Exception', self.__message_type, self.__message_token)

    def append_summary_log(self, content):
        # 打印并加入到最终日志
        # date_str = DateFormatter.convert_local_date_to_string(datetime.now(), '%H:%M:%S')
        self.record_simple_info(content)
        self.__summary_log = self.__summary_log + content + '\n\n'

    def record_summary_log(self, content=None):
        if content is not None:
            self.append_summary_log(content)
        MessageHandler.send_message(self.__summary_log, 'Procedure Summary', self.__message_type, self.__message_token)
        self.__summary_log = ""



import math
import pandas as pd
import time
from datetime import datetime, timedelta
from ..generic.spot_fetching import *
from ..util.logger import ProcedureRecorder
from cy_data_access.models.position import *
from cy_widgets.trader.exchange_trader import *
from cy_widgets.logger.trading import *
from cy_widgets.strategy.exchange.aims import *
from cy_widgets.exchange.provider import ExchangeType


class BinanceAIMS:
    """AIMS Procedure
    Require: 连接 DB_POSITION 数据库
    """

    def __init__(self,
                 coin_pair: CoinPair,
                 time_frame,
                 provider: CCXTProvider,
                 signal_scale,
                 ma_periods,
                 sell_threshold,
                 invest_base_amount,
                 recorder: ProcedureRecorder,
                 fee_percent=0,  # 订单里已经扣了，这里不用手动算
                 debug=False):
        self.__df = pd.DataFrame()
        self.__provider = provider
        self.__recorder = recorder
        self.__coin_pair = coin_pair
        self.__recorder.append_summary_log(
            '**{} - {} - AIMS**'.format(provider.display_name.upper(), coin_pair.formatted().upper()))
        self.__recorder.append_summary_log(
            '**日期**: {} \n'.format(DateFormatter.convert_local_date_to_string(datetime.now(), "%Y-%m-%d")))
        self.__fee_percent = fee_percent
        # 直接从远往近抓
        self.__configuration = ExchangeFetchingConfiguration(
            coin_pair, time_frame, 3, ExchangeFetchingType.FILL_RECENTLY, debug=debug)
        # 策略参数
        self.__ma_periods = ma_periods
        self.__signal_scale = signal_scale
        self.__sell_threshold = sell_threshold
        # 交易
        self.__invest_base_amount = invest_base_amount

    def run_task(self):
        try:
            while True:
                self.fetch_candle()
                # 循环，直到最后一条数据是今天的
                if self.__df.shape[0] > 0 and self.__df.iloc[-1][COL_CANDLE_BEGIN_TIME].dayofweek == datetime.utcnow().weekday():
                    break
                self.__recorder.record_exception("未抓取到今天的 K 线，稍后重试")
                time.sleep(3)
            invest_ratio = self.calculate_signal()
            self.handle_signal(invest_ratio)
        except Exception as e:
            self.__recorder.record_summary_log(str(e))

    def fetch_candle(self):
        """ 获取 K 线 """
        def __get_latest_date():
            if self.__df.shape[0] > 0:
                return self.__df[COL_CANDLE_BEGIN_TIME].iloc[-1]
            return datetime.now() - timedelta(days=self.__ma_periods + 15)  # 往前多加15天开始

        def __save_df(data_df: pd.DataFrame):
            before_count = self.__df.shape[0]

            # 为了去重
            data_df.set_index(COL_CANDLE_BEGIN_TIME, inplace=True)
            if before_count > 0:
                self.__df.set_index(COL_CANDLE_BEGIN_TIME, inplace=True)
            df_res = pd.concat([self.__df, data_df[~data_df.index.isin(self.__df.index)]])
            df_res.update(data_df)

            self.__df = df_res
            # 排序后重置 index
            self.__df.sort_index(inplace=True)
            self.__df.reset_index(inplace=True)

            after_count = self.__df.shape[0]
            # 前后数量不一致，说明有新数据，继续尝试获取
            return before_count != after_count

        procedure = ExchangeFetchingProcedure(ExchangeFetcher(self.__provider),
                                              self.__configuration,
                                              None,
                                              __get_latest_date,
                                              __save_df)

        # Fill to Latest
        procedure.run_task()

        if self.__configuration.debug:
            print(self.__df)

        self.__recorder.record_procedure("获取 K 线成功")

    @property
    def __position(self):
        return AIMSPosition.position_with(self.__provider.display_name, self.__coin_pair.formatted())

    @property
    def __trader_logger(self):
        return TraderLogger(self.__provider.display_name, self.__coin_pair.formatted(), 'Spot', self.__recorder)

    def calculate_signal(self):
        # Signal Calculation
        strategy = AutoInvestVarietalStrategy(signal_scale=self.__signal_scale,
                                              ma_periods=self.__ma_periods)
        signals = strategy.calculate_signals(self.__df, False)
        # Debug 下打印一点信号看看
        if self.__configuration.debug:
            print(signals[-200:])
        return signals.iloc[-1][COL_SIGNAL]

    def handle_signal(self, signal):
        """处理信号"""
        if signal > 0:
            self.__recorder.append_summary_log('**信号**: 买入({}) \n'.format(signal))
            # 买入
            self.handle_buying(signal * self.__invest_base_amount)
        else:
            position = self.__position
            open_price = self.__df.iloc[-1][COL_OPEN]
            # P/L
            pol = open_price / (position.cost / position.hold) if position.hold > 0 else 0
            if pol > self.__sell_threshold:
                # 用数据库判断卖出，这里再取实际的仓位数
                amount = self.__provider.ccxt_object_for_order.fetch_balance()['free'][self.__coin_pair.trade_coin]
                # 卖出
                self.handle_selling(amount)
            else:
                # 无信号 结束
                self.__recorder.append_summary_log('**信号**: 无 \n')
                if pol > 0:
                    msg = """**持仓数量**: {} {} \n
**仓位成本**: {} {} \n
**仓位均价**: {} \n
**P/L**: {}%
""".format(round(position.hold, 8), self.__coin_pair.trade_coin.upper(),
                        round(position.cost, 8), self.__coin_pair.base_coin.upper(),
                        round((position.cost / position.hold) if position.hold > 0 else 0, 8), round(100*pol, 2))
                    self.__recorder.append_summary_log(msg)
                self.__recorder.record_summary_log()

    def handle_buying(self, amount):
        # 下单数量
        self.__recorder.append_summary_log('**下单数量**: {} {}\n'.format(amount, self.__coin_pair.base_coin.upper()))
        # ers
        order = Order(self.__coin_pair, amount, 0)  # Only set base coin amount
        executor = ExchangeOrderExecutorFactory.executor(self.__provider, order, self.__trader_logger)
        # place order
        response = executor.handle_long_order_request()
        if response is None:
            self.__recorder.record_summary_log('**下单失败**')
            return
        # handle order info
        price = response['average']
        cost = response['cost']
        filled = response['filled']
        # Binance 手续费已经扣掉了
        # {'id': '701037299',
        # 'clientOrderId': 'Se8IFlpHyWpsY7OaYkhKC1',
        # 'timestamp': 1597589153872,
        # 'datetime': '2020-08-16T14:45:53.872Z',
        # 'lastTradeTimestamp': None,
        # 'symbol': 'BNB/USDT',
        # 'type': 'limit',
        # 'side': 'buy',
        # 'price': 23.3236,
        # 'amount': 0.47,
        # 'cost': 10.854319,
        # 'average': 23.094295744680853,
        # 'filled': 0.47,
        # 'remaining': 0.0,
        # 'status': 'closed',
        # 'fee': None,
        # 'trades': None}
        buy_amount = math.floor(filled * (1 - self.__fee_percent) * 1e8) / 1e8  # *1e8 向下取整再 / 1e8
        msg = """**成交价格**: {} \n
**成交总价**: {} {}\n
**成交数量**: {} {}
""".format(round(price, 6), round(cost, 6), self.__coin_pair.base_coin.upper(), round(buy_amount, 8), self.__coin_pair.trade_coin.upper())
        self.__recorder.append_summary_log(msg)
        # 更新 Cost/Hold 到数据库
        position = self.__position
        position.update(cost, buy_amount)
        msg = """**持仓数量**: {} {} \n
**仓位成本**: {} {} \n
**仓位均价**: {}
""".format(round(position.hold, 8), self.__coin_pair.trade_coin.upper(),
           round(position.cost, 8), self.__coin_pair.base_coin.upper(),
           round((position.cost / position.hold) if position.hold > 0 else 0, 8))
        self.__recorder.record_summary_log(msg)

    def handle_selling(self, amount):
        # 下单数量
        self.__recorder.append_summary_log('**信号**: 卖出 \n')
        self.__recorder.append_summary_log('**下单数量**: {} {}\n'.format(amount, self.__coin_pair.trade_coin.upper()))
        # ers
        order = Order(self.__coin_pair, 0, amount, side=OrderSide.SELL)  # Only set trade coin amount
        executor = ExchangeOrderExecutorFactory.executor(self.__provider, order, self.__trader_logger)
        # place order
        response = executor.handle_close_order_request()
        if response is None:
            self.__recorder.record_summary_log('**下单失败**')
            return
        # handle order
        # 'id': '701205190',
        # 'clientOrderId': 'Iep2bjMRJGSUBS0qDq9AhS',
        # 'timestamp': 1597595099771,
        # 'datetime': '2020-08-16T16:24:59.771Z',
        # 'lastTradeTimestamp': None,
        # 'symbol': 'BNB/USDT',
        # 'type': 'limit',
        # 'side': 'sell',
        # 'price': 22.979,
        # 'amount': 0.92,
        # 'cost': 21.355224,
        # 'average': 23.2122,
        # 'filled': 0.92,
        # 'remaining': 0.0,
        # 'status': 'closed',
        # 'fee': None,
        # 'trades': None}
        price = response['average']
        cost = response['cost']
        filled = response['filled']
        position = self.__position
        msg = """**成交价格**: {} \n
**成交总价**: {} {}\n
**成交数量**: {} {}\n
**仓位成本**: {} {}
""".format(round(price, 6), round(cost, 6), self.__coin_pair.base_coin.upper(), round(filled, 8), self.__coin_pair.trade_coin.upper(),
           round(position.cost, 8), self.__coin_pair.base_coin.upper())
        # 更新 Cost/Hold 到数据库
        position.reset()
        self.__recorder.record_summary_log(msg)



from multiprocessing.pool import Pool
from datetime import datetime
import numpy as np
import time


class MultiTest:
    n = 0

    def print_something(self, param):
        print(self.n)
        self.n += 1
        return np.math.factorial(param)

    def run_task(self):
        start_time = datetime.now()
        params = np.arange(0, 10000)
        with Pool(processes=4) as pool:
            res = pool.map(self.print_something, params)
            print(datetime.now() - start_time)


MultiTest().run_task()



import os
from datetime import datetime
from cy_components.helpers.formatter import CandleFormatter
from cy_data_access.models.market import *
from cy_data_access.util.convert import convert_df_to_json_list
from ..generic.spot_fetching import *


class CSVHistoricalSpotCandle:
    def __init__(self, coin_pair, time_frame, exchange_type, start_date='2020-03-03 00:00:00', end_date=None, save_path=None, per_limit=1000):
        self.__df = pd.DataFrame()
        self.save_path = save_path
        # 时间区间
        self.start_date = DateFormatter.convert_string_to_local_date(start_date).astimezone()
        self.end_date = DateFormatter.convert_string_to_local_date(end_date) if end_date is not None else datetime.now()
        self.end_date = self.end_date.astimezone()
        # 抓取使用的配置
        self.provider = CCXTProvider("", "", exchange_type)
        self.config = ExchangeFetchingConfiguration(
            coin_pair, time_frame, 1, ExchangeFetchingType.FILL_RECENTLY, batch_limit=per_limit)
        self.fetcher = ExchangeFetcher(self.provider)

    def run_task(self):
        # 开始抓取
        ExchangeFetchingProcedure(self.fetcher, self.config, None, self.__get_latest_date, self.__save_df).run_task()
        return self.__df

    def __get_latest_date(self):
        if self.__df.shape[0] > 0:
            return self.__df[COL_CANDLE_BEGIN_TIME].iloc[-1]
        # 从开始时间，往最近的抓
        return self.start_date

    def __save_df(self, data_df: pd.DataFrame):
        before_count = self.__df.shape[0]

        # 为了去重
        data_df.set_index(COL_CANDLE_BEGIN_TIME, inplace=True)
        if before_count > 0:
            self.__df.set_index(COL_CANDLE_BEGIN_TIME, inplace=True)
        df_res = pd.concat([self.__df, data_df[~data_df.index.isin(self.__df.index)]])
        df_res.update(data_df)

        self.__df = df_res
        # 排序后重置 index
        self.__df.sort_index(inplace=True)
        self.__df.reset_index(inplace=True)

        after_count = self.__df.shape[0]
        # 没有新增
        hasnt_new_candle = before_count == after_count
        is_reach_end_date = self.__df[COL_CANDLE_BEGIN_TIME].iloc[-1] >= self.end_date
        stop = hasnt_new_candle or is_reach_end_date
        if stop:
            self.__df = self.__df[(self.__df[COL_CANDLE_BEGIN_TIME] >= self.start_date)
                                  & (self.__df[COL_CANDLE_BEGIN_TIME] <= self.end_date)]
            # self.__df.set_index(COL_CANDLE_BEGIN_TIME, inplace=True)
            self.__df.reset_index(inplace=True, drop=True)
            # print(self.__df)
            if self.save_path is not None:
                self.__save_df_to_csv()
        return not stop

    def __save_df_to_csv(self):
        path = "{}/{}".format(self.save_path, self.provider.ccxt_object_for_fetching.id)
        os.path.join(path)
        if not os.path.exists(path):
            os.makedirs(path, exist_ok=True)
        path = "{}/{}-{}.csv".format(path, self.config.coin_pair.formatted('-'), self.config.time_frame.value.upper())
        self.__df.to_csv(path)


class DBHistoricalSpotCandle:
    """外部连接数据库，内部只负责写入"""

    def __init__(self, coin_pair, time_frame, exchange_type, start_date='2020-03-03 00:00:00', end_date=None, per_limit=1000):
        # 时间区间
        self.start_date = DateFormatter.convert_string_to_local_date(start_date).astimezone()
        self.end_date = DateFormatter.convert_string_to_local_date(end_date) if end_date is not None else datetime.now()
        self.end_date = self.end_date.astimezone()
        # 抓取使用的配置
        self.provider = CCXTProvider("", "", exchange_type)
        self.config = ExchangeFetchingConfiguration(
            coin_pair, time_frame, 1, ExchangeFetchingType.FILL_RECENTLY, batch_limit=per_limit)
        self.fetcher = ExchangeFetcher(self.provider)
        # table name
        self.candle_cls = candle_record_class_with_components(self.provider.display_name, coin_pair, time_frame)

    def run_task(self):
        # 开始抓取
        ExchangeFetchingProcedure(self.fetcher, self.config, None, self.__get_latest_date, self.__save_df).run_task()

    def __get_latest_date(self):
        return self.start_date

    def __save_df(self, df: pd.DataFrame):
        if df.shape[0] == 0:
            return False
        # 最后日期
        self.start_date = df.sort_values(COL_CANDLE_BEGIN_TIME, ascending=False)[COL_CANDLE_BEGIN_TIME].iloc[0]
        # 保存
        json_list = convert_df_to_json_list(df, COL_CANDLE_BEGIN_TIME)
        self.candle_cls.bulk_upsert_records(json_list)
        return self.start_date < self.end_date



"""Top-level package for CY Procedure."""
try:
    __import__('pkg_resources').declare_namespace(__name__)
except ImportError:
    __path__ = __import__('pkgutil').extend_path(__path__, __name__)

__author__ = """CY Gatro"""
__email__ = 'cragodn@gmail.com'
__version__ = '0.3.0'



import math
import pandas as pd
from datetime import datetime, timedelta
from ..generic.spot_fetching import *
from ..util.logger import ProcedureRecorder
from cy_widgets.trader.exchange_trader import *
from cy_widgets.logger.trading import *
from cy_widgets.strategy.exchange.autobuy import *
from cy_widgets.exchange.provider import ExchangeType


class OKexAIP:
    """https://www.okex.com/docs/zh/#error-spot"""

    def __init__(self,
                 coin_pair: CoinPair,
                 time_frame,
                 signal_provider,
                 day_of_week,
                 ma_periods,
                 trader_provider,
                 invest_base_amount,
                 recorder: ProcedureRecorder,
                 debug=False,
                 fee_percent=0.0015):
        """余币宝转出USDT + 定投买入 + 余币宝转入BTC

        Parameters
        ----------
        coin_pair : CoinPair
            币对信息
        time_frame : TimeFrame
            K线间隔
        signal_provider : CCXTProvider
            抓取K线
        interval_days : Int
            计算定投信号用，间隔天数
        ma_periods : Int
            计算定投信号用，MA计算参数
        start_index : Int
            计算定投信号用，相对星期
        trader_provider : CCXTProvider
            交易用
        invest_base_amount : Double
            定投基准额
        recorder : ProcedureRecorder
            记录
        fee_percent : Double
            手续费
        """
        self.__df = pd.DataFrame()
        self.signal_provider = signal_provider
        self.trader_provicer = trader_provider
        self.recorder = recorder
        self.coin_pair = coin_pair
        self.recorder.append_summary_log('**{} - 定投**'.format(coin_pair.formatted().upper()))
        self.fee_percent = fee_percent
        # 直接从远往近抓
        self.configuration = ExchangeFetchingConfiguration(
            coin_pair, time_frame, 3, ExchangeFetchingType.FILL_RECENTLY, debug=debug)
        # 策略参数
        self.day_of_week = datetime.today().weekday()
        self.ma_periods = ma_periods
        self.invest_day = day_of_week
        # 交易
        self.__invest_base_amount = invest_base_amount

    def run_task(self):
        self.__fetch_candle()
        invest_ratio = self.__calculate_signal()
        if invest_ratio is not None and invest_ratio > 0 and self.invest_day == self.day_of_week:
            self.__place_invest_order(invest_ratio)
        else:
            self.recorder.record_summary_log()

    def __fetch_candle(self):
        procedure = ExchangeFetchingProcedure(ExchangeFetcher(self.signal_provider),
                                              self.configuration,
                                              None,
                                              self.__get_latest_date,
                                              self.__save_df)

        # Fill to Latest
        procedure.run_task()

        if self.configuration.debug:
            print(self.__df)

        self.recorder.record_procedure("获取 K 线成功")

    def __calculate_signal(self):
        # Calculation
        strategy = AutoBuyCoinStrategy(day_of_week=self.day_of_week,
                                       ma_periods=self.ma_periods)
        signals = strategy.calculate_signals(self.__df, False)
        if self.configuration.debug:
            print(signals[-200:])
        date_string = DateFormatter.convert_local_date_to_string(
            signals['candle_begin_time'][self.__df.index[-1]], "%Y-%m-%d")
        actual_ratio = signals['high_change'][self.__df.index[-1]]
        advance_ratio = signals['signal'][self.__df.index[-1]]
        close_price = signals['close'][self.__df.index[-1]]
        msg = """**日期**: {} \n
**计算价格**: {} \n
**市场信号**: {} \n
**定投倍数**: {}
""".format(date_string, round(close_price, 4), round(actual_ratio, 4), advance_ratio)
        self.recorder.append_summary_log(msg)
        return advance_ratio

    def __place_invest_order(self, ratio):
        # 实际定投数
        invest_amount = ratio * self.__invest_base_amount
        self.recorder.append_summary_log('**定投额**({} * {}): {} {}'.format(self.__invest_base_amount,
                                                                          ratio, invest_amount, self.coin_pair.base_coin.upper()))
        self.__invest_proccess(invest_amount)

    def __invest_proccess(self, invest_amount):
        # ybb->bb
        if not self.__transfer_amount(invest_amount, self.coin_pair.base_coin.lower(), 8, 1):
            self.recorder.record_summary_log('**划转{}失败**'.format(self.coin_pair.base_coin.upper()))
            return
        # balance
        balance = self.trader_provicer.ccxt_object_for_fetching.fetch_balance()
        usdt = balance['free']['USDT']
        # not enough
        if usdt < invest_amount:
            self.recorder.record_summary_log('**{}不足以定投**'.format(self.coin_pair.base_coin.upper()))
        # place order
        logger = TraderLogger(self.trader_provicer.display_name, self.coin_pair.formatted(), 'Spot', self.recorder)
        order = Order(self.coin_pair, invest_amount, 0)
        executor = ExchangeOrderExecutorFactory.executor(self.trader_provicer, order, logger)
        """{'info': {'client_oid': '',
            'created_at': '2020-06-25T07:52:44.171Z',
            'filled_notional': '9.900909395',
            'filled_size': '0.00107953',
            'funds': '',
            'instrument_id': 'BTC-USDT',
            'notional': '',
            'order_id': '5133799533470720',
            'order_type': '0',
            'price': '9263.2',
            'price_avg': '9171.5',
            'product_id': 'BTC-USDT',
            'side': 'buy',
            'size': '0.00107953',
            'state': '2',
            'status': 'filled',
            'timestamp': '2020-06-25T07:52:44.171Z',
            'type': 'limit'},
            'id': '5133799533470720',
            'clientOrderId': '',
            'timestamp': 1593071564171,
            'datetime': '2020-06-25T07:52:44.171Z',
            'lastTradeTimestamp': None,
            'symbol': 'BTC/USDT',
            'type': 'limit',
            'side': 'buy',
            'price': 9263.2,
            'average': 9171.5,
            'cost': 9.900909395,
            'amount': 0.00107953,
            'filled': 0.00107953,
            'remaining': 0,
            'status': 'closed',
            'fee': None,
            'trades': None}"""
        response = executor.handle_long_order_request()
        if response is None:
            self.__transfer_amount(invest_amount, self.coin_pair.base_coin.lower(), 1, 8)
            self.recorder.record_summary_log('**下单失败**')
            return
        # order info
        price = response['average']
        cost = response['cost']
        buy_amount = math.floor(response['filled'] * (1 - self.fee_percent) * 1e8) / 1e8  # *1e8 向下取整再 / 1e8
        msg = """**下单价格**: {} \n
**下单总价**: {} \n
**买入数量**: {}
""".format(round(price, 6), round(cost, 6), round(buy_amount, 8))
        self.recorder.append_summary_log(msg)
        # bb->ybb (RemainingBaseCoin)
        remaining_base_coin = round(invest_amount - cost, 6)  # 保留 6 位
        if remaining_base_coin > 0 and not self.__transfer_amount(remaining_base_coin, self.coin_pair.base_coin.lower(), 1, 8):
            self.recorder.record_summary_log('**划转{}失败**'.format(self.coin_pair.base_coin.upper()))
            return
        # bb->ybb (InvestCoin)
        if not self.__transfer_amount(buy_amount, self.coin_pair.trade_coin.lower(), 1, 8):
            self.recorder.record_summary_log('**划转{}失败**:{}'.format(self.coin_pair.trade_coin.upper(), order))
            return
        self.recorder.record_summary_log('**定投成功**')

    def __transfer_amount(self, amount, coin, from_type, to_type):
        """ybb和bb互转

        Parameters
        ----------
        coin : str
            usdt/btc/...
        amount : double
            数量
        from_type : int
        to_type : int
            1: bb 8: ybb
        """
        # ybb -> bb
        try:
            response = self.trader_provicer.ccxt_object_for_order.account_post_transfer({
                'currency': coin,
                'amount': '{}'.format(amount),
                'type': '0',
                'from': '{}'.format(from_type),
                'to': '{}'.format(to_type)
            })
            return response['result']
        except Exception as e:
            self.recorder.append_summary_log('{}.{}.划转失败.{}'.format(from_type, to_type, str(e)))
            return False

    def __get_latest_date(self):
        if self.__df.shape[0] > 0:
            return self.__df[COL_CANDLE_BEGIN_TIME].iloc[-1]
        return datetime.now() - timedelta(days=self.ma_periods + 15)  # 往前多加15天开始

    def __save_df(self, data_df: pd.DataFrame):
        before_count = self.__df.shape[0]

        # 为了去重
        data_df.set_index(COL_CANDLE_BEGIN_TIME, inplace=True)
        if before_count > 0:
            self.__df.set_index(COL_CANDLE_BEGIN_TIME, inplace=True)
        df_res = pd.concat([self.__df, data_df[~data_df.index.isin(self.__df.index)]])
        df_res.update(data_df)

        self.__df = df_res
        # 排序后重置 index
        self.__df.sort_index(inplace=True)
        self.__df.reset_index(inplace=True)

        after_count = self.__df.shape[0]
        # 前后数量不一致，说明有新数据，继续尝试获取
        return before_count != after_count
